// Вспомогательная функция для создания трехмерного вектора
function vec3(x, y, z, u = 0, v = 0) {

    return {
        x: x, y: y, z: z,
        u: u, v: v
    };
}

class Caster3 {

    constructor(w = 320, h = 200) {

        this._width  = w;
        this._height = h;
        this._buffer_c = new Uint32Array(w * h);
    }

    // Для импорта
    get_buffer(type = 0) {
        return this._buffer_c;
    }

    // Быстрая очистка памяти
    cls(cl = 0) {
        this._buffer_c.fill(cl);
    }

    // Установка цвета в буфер
    pset(x, y, cl) {

        if (x >= 0 && y >= 0 && x < this._width && y < this._height) {
            this._buffer_c[y*this._width + x] = cl;
        }
    }

    // Извлечение цвета из буфера
    point(x, y) {

        if (x >= 0 && y >= 0 && x < this._width && y < this._height) {
            return this._buffer_c[y*this._width + x];
        }

        return 0;
    }

    // Растеризация уже спроецированного треугольника
    raster(vtx, param = null) {

        // Проверка видимости грани
        let AB = {x: vtx[1].x - vtx[0].x, y: vtx[1].y - vtx[0].y};
        let AC = {x: vtx[2].x - vtx[0].x, y: vtx[2].y - vtx[0].y};

        // Отбросить грань при обходе против часовой стрелки
        if (AB.x*AC.y - AB.y*AC.x <= 0) {
            return;
        }

        // Цвет для рисования
        let color = param ? (param.color || 0xFFFFFF) : 0xFFFFFF;

        // Текстурирование и точки пересечения с треугольником
        let R   = (param && typeof param.R === 'object') ? param.R : null;
        let TXT = (param && typeof param.T === 'object') ? param.T : null;
        let TM  = null;

        // Вычисления половин экрана
        let [W2, H2] = [this._width >> 1, this._height >> 1];

        // Подготовка текстуры
        if (R) {
            TM = {
                u: 0, uz: R.U.z*H2,
                v: 0, vz: R.V.z*H2,
                d: 0, dz: R.D.z*H2,
            };
        }

        // Сортировка по возрастанию
        let p = vtx.sort(function (x, y) { return x.y < y.y ? -1 : 1; });

        // Главная сторона
        let [Ax,  Adx] = [p[0].x, 0];
        let [ACx, ACy] = [p[2].x - p[0].x, p[2].y - p[0].y];

        // Два треугольника
        for (let i = 0; i < 2; i++) {

            // Боковая сторона
            let [Bx,  Bdx] = [p[i].x, 0];
            let [AMx, AMy] = [p[i+1].x - p[i].x, p[i+1].y - p[i].y];

            // Отрисовка полутреугольника
            for (let y = p[i].y; y < p[i+1].y + i; y++) {

                // Находится за пределами
                if (y >= this._height) break;

                // Рисовать только если не за верхней границей экрана
                if (y >= 0) {

                    let x1 = Math.min(Ax, Bx);
                    let x2 = Math.max(Ax, Bx);

                    // Концы отрезков не выходят за пределы рисования
                    if (x1 < this._width || x2 >= 0) {

                        x1 = Math.max(x1, 0);
                        x2 = Math.min(x2, this._width - 1);

                        // Запущена растеризация с текстурой
                        if (R) {

                            // Направление камеры всегда стандартное
                            let d = {x: x1 - W2, y: H2 - y};

                            // Рассчитать новую позицию перед растеризаицией линии
                            TM.u = R.U.x*d.x + R.U.y*d.y + TM.uz;
                            TM.v = R.V.x*d.x + R.V.y*d.y + TM.vz;
                            TM.d = R.D.x*d.x + R.D.y*d.y + TM.dz;
                        }

                        // Прочертить линию
                        for (let x = x1; x <= x2; x++) {

                            // Стандартный цвет
                            let cl = color, u, v;

                            // Вычисление позиции (u,v) для (tx,ty)
                            if (R) {

                                [u, v] = [TM.u / TM.d, TM.v / TM.d];

                                if (TXT) {

                                    // Рассчитать позицию текстуры позицию
                                    let tx = R.tu.x + R.tu.y*u + R.tu.z*v;
                                    let ty = R.tv.x + R.tv.y*u + R.tv.z*v;

                                    cl = TXT.point(tx, ty);
                                }
                            }

                            // Рисование точки в буфере
                            this.pset(x, y, cl);

                            // Приращение +1
                            if (R) { TM.u += R.U.x; TM.v += R.V.x; TM.d += R.D.x; }
                        }
                    }
                }

                // Сдвиг точек x1,x2. Важно использовать trunc!
                Adx += ACx; Ax += Math.trunc(Adx / ACy); Adx %= ACy;
                Bdx += AMx; Bx += Math.trunc(Bdx / AMy); Bdx %= AMy;
            }
        }
    }

    // Создать треугольник по трем точкам (array a[3], array b[3], array c[3])
    create3(a, b, c) {

        return [
            vec3(a[0], a[1], a[2], a[3] || 0, a[4] || 0),
            vec3(b[0], b[1], b[2], b[3] || 0, b[4] || 0),
            vec3(c[0], c[1], c[2], c[3] || 0, c[4] || 0),
        ];
    }

    // Создать текстурный объект
    texture_make(w, h) {

        return {
            w: w,
            h: h,
            data: new Uint32Array(w * h),
            pset: function(x, y, cl) {
                if (x >= 0 && y >= 0 && x < this.w && y < this.h)
                    this.data[this.w*Math.trunc(y) + Math.trunc(x)] = cl;
            },
            point: function(x, y) {
                if (x < 0) x = 0; else if (x >= this.w) x = this.w-1;
                if (y < 0) y = 0; else if (y >= this.h) x = this.h-1;
                return this.data[this.w*Math.trunc(y) + Math.trunc(x)];
            }
        };
    }

    // Вычисление параметров треугольника:
    // T (create) - создан с помощью метода create
    // P (pivor) - точка отсчета (камера)
    compute(T, P = null) {

        P = (P === null) ? vec3(0,0,0) : P;

        let AB = vec3(T[1].x - T[0].x, T[1].y - T[0].y, T[1].z - T[0].z);
        let AC = vec3(T[2].x - T[0].x, T[2].y - T[0].y, T[2].z - T[0].z);
        let AP = vec3(   P.x - T[0].x,    P.y - T[0].y,    P.z - T[0].z);

        return {

            // Параметры
            U: vec3(
                AC.y * AP.z - AC.z * AP.y,
                AC.z * AP.x - AC.x * AP.z,
                AC.x * AP.y - AC.y * AP.x
            ),
            V: vec3(
                AB.z * AP.y - AB.y * AP.z,
                AB.x * AP.z - AB.z * AP.x,
                AB.y * AP.x - AB.x * AP.y
            ),
            D: vec3(
                AB.z * AC.y - AB.y * AC.z,
                AB.x * AC.z - AB.z * AC.x,
                AB.y * AC.x - AB.x * AC.y
            ),

            // Вектор нормали
            N: vec3(
                AB.y * AC.z - AB.z * AC.y,
                AB.z * AC.x - AB.x * AC.z,
                AB.x * AC.y - AB.y * AC.x
            ),

            // Текстурные координаты
            tu: vec3(T[0].u, T[1].u - T[0].u, T[2].u - T[0].u),
            tv: vec3(T[0].v, T[1].v - T[0].v, T[2].v - T[0].v),

            // Сохранить треугольник для перевычислений
            a: T[0], AB: AB, AC: AC, AP: AP,
        };
    }

    // Обновление стартовой позиции камеры для вычисленного треугольника
    // R - объект описания (результат compute)
    // P - новая точка отсчета
    update3(R, P) {

        let AP = vec3(P.x - R.a.x, P.y - R.a.y, P.z - R.a.z);
        let [AB, AC] = [R.AB, R.AC];

        R.U = vec3(
            AC.y * AP.z - AC.z * AP.y,
            AC.z * AP.x - AC.x * AP.z,
            AC.x * AP.y - AC.y * AP.x
        );

        R.V = vec3(
            AB.z * AP.y - AB.y * AP.z,
            AB.x * AP.z - AB.z * AP.x,
            AB.y * AP.x - AB.x * AP.y
        );

        R.AP = AP;
    }

    // Проецирование массива векторов типа vec3
    project(v) {

        let out = [];
        let ppd = this._height / 2;

        for (let i = 0; i < v.length; i++) {

            out.push({
                x: Math.floor(this._width  / 2 + ppd*v[i].x / v[i].z),
                y: Math.floor(this._height / 2 - ppd*v[i].y / v[i].z),
            });
        }

        return out;
    }

    // РАБОТА С МАТРИЦАМИ
    // =================================================================

    // Единичная матрица
    identity() {

        return [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ];
    }

    // Умножение матриц 4x4
    mult(m1, m2) {

        let mat = [];
        for (let a = 0; a < 4; a++) {

            let row = [];
            for (let b = 0; b < 4; b++) {

                let s = 0;
                for (let c = 0; c < 4; c++) s += m1[a][c]*m2[c][b];
                row.push(s);
            }

            mat.push(row);
        }

        return mat;
    }

    // Поворот по оси X
    rotateX(m, angle) {

        let [s,c] = [-Math.sin(angle), Math.cos(angle)];

        return this.mult(m, [
            [1, 0,  0, 0],
            [0, c, -s, 0],
            [0, s,  c, 0],
            [0, 0,  0, 1]
        ]);
    }

    // Поворот по оси Y
    rotateY(m, angle) {

        let [s,c] = [-Math.sin(angle), Math.cos(angle)];

        return this.mult(m, [
            [ c, 0, s, 0],
            [ 0, 1, 0, 0],
            [-s, 0, c, 0],
            [ 0, 0, 0, 1]
        ]);
    }

    // Поворот по оси Z
    rotateZ(m, angle) {

        let [s,c] = [Math.sin(angle), Math.cos(angle)];

        return this.mult(m, [
            [c, -s, 0, 0],
            [s,  c, 0, 0],
            [0,  0, 1, 0],
            [0,  0, 0, 1]
        ]);
    }

    // Трансляция матрицы
    translate(m, p) {

        return this.mult(m, [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [p.x, p.y, p.z, 1]
        ]);

    }

    // vtx - массив векторов для выполнения преобразования матрицы
    apply_matrix(vtx, m) {

        let out = [];
        for (let i = 0; i < vtx.length; i++) {

            let t = [vtx[i][0], vtx[i][1], vtx[i][2], 1];
            let n = [0, 0, 0];
            for (let a = 0; a < 3; a++) {

                let s = 0;
                for (let b = 0; b < 4; b++) s += t[b] * m[b][a];
                n[a] = s;
            }

            out.push( { x:n[0], y:n[1], z:n[2] } );
        }

        return out;
    }
}
